(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                       Calcul parallle de fractals                     |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Printf
open Parallele
open Global
open Widgets
open Graphics
open Calc
open Param
open Draw


(* gestion du signal SIGPIPE *)
exception SIGPIPE
let handle_sigpipe _ = raise SIGPIPE
let _ = Sys.signal Sys.sigpipe (Sys.Signal_handle handle_sigpipe)


                   (* +-------------------------+
                      |  Slectionne un onglet  |
                      +-------------------------+ *)

let sel_onglet data titre =
  while (List.hd data.ongl).titre <> titre do
    data.ongl <- List.tl(data.ongl) @ [List.hd(data.ongl)]
  done


                 (* +-----------------------------+
                    |  Redimensionne l'affichage  |
                    +-----------------------------+ *)

let resize data =

  wx := size_x();
  wy := size_y();

  let x0 = 0                   in
  let x1 = x0 + pad            in
  let x2 = x1 + pad            in
  let x3 = x2 + fl_width + 2   in
  let x4 = x3 + 2              in
  let x9 = max (size_x()) minwidth in
  let x8 = x9 - pad            in
  let x7 = x8 - pad            in
  let x6 = x7 - fl_width - 2   in
  let x5 = x6 - 2              in
  let sx = x5 - x4             in

  let y0 = 0                   in
  let y1 = y0 + pad            in
  let y2 = y1 + pad            in
  let y3 = y2 + bheight + 1    in
  let y4 = y3 + fl_width + 2   in
  let y5 = y4 + 2              in
  let y9 = (max (size_y()) minheight) - 4 in
  let y8 = y9 - oheight - pad  in
  let y7 = y8 - fl_width - 2   in
  let y6 = y7 - 2              in
  let sy = y6 - y5             in

  let fonglets = {fx=x1; fy=y1; fw=x8-x1; fh=y9-y1}
  and fdraw    = {fx=x4; fy=y5; fw=x5-x4; fh=y6-y5}
  and fmsg     = {fx=x2; fy=y2; fw=x7-x2; fh=bheight}
  and titre    = (List.hd data.ongl).titre in
  let ongl     = make_onglets fonglets
		  ["Paramtres"; "Mandelbrot"; "Julia"; "Cache"] in

  let set_size img = {
    img with
      dx = img.dx*.float(sx)/.float(img.sx);
      dy = img.dy*.float(sy)/.float(img.sy);
      sx = sx;
      sy = sy
  } in

  data.julia  <- set_size data.julia;
  data.mandel <- set_size data.mandel;
  data.mem    <- [];
  data.ongl   <- ongl;
  data.fongl  <- fonglets;
  data.fdraw  <- fdraw;
  data.fmsg   <- fmsg;
  sel_onglet data titre;

  set_color fond_gris;
  fill_rect 0 0 !wx !wy


                               (* +----------+
                                  |  Matre  |
                                  +----------+ *)

let maitre () =

  let port   = ref(0)
  and slaves = ref []
  and help   = ref false
  and auto   = ref false
  and maxmem = ref(10)
  in

  let (=::) l x = l := x :: !l in

  let argspec = [
    "-cache", Arg.Int((:=) maxmem), "<int> nombre de dessins conservs en mmoire";
    "-p", Arg.Int((:=) port),       "<int> port TCP";
    "-s", Arg.String((=::) slaves), "<string:int> esclave  contacter";
    "-auto", Arg.Set(auto),         "mode autonome";
    "-v", Arg.Set(debug),      	    "trace les communications"]

  and anon(s) = raise (Arg.Bad ("paramtre non reconnu : " ^ s))
  and msg ="syntaxe : julia maitre <paramtres>"
  in

  Arg.current := 1;
  Arg.parse argspec anon msg;
  
  if !help then begin
    Arg.usage argspec msg;
    exit(1)
  end;

  let image_julia = {
    nat = JULIA;
    x0 = 0.0; 	   y0 = 0.0;
    dx = 3.0;      dy = 2.0;
    sx = 600;      sy = 400;
    jx = -0.7998;  jy = 0.15517;
    hx = 200;      hy = 20;
    nmax  = 200;
    psize = 32
  } in
  let image_mandelbrot = { image_julia with nat = MANDELBROT; x0 = -0.5 } in

  let x0 = 0                   in
  let x1 = x0 + pad            in
  let x2 = x1 + pad            in
  let x3 = x2 + fl_width + 2   in
  let x4 = x3 + 2              in
  let x5 = x4 + image_julia.sx in
  let x6 = x5 + 2              in
  let x7 = x6 + fl_width + 2   in
  let x8 = x7 + pad            in
  let x9 = x8 + pad            in

  let y0 = 0                   in
  let y1 = y0 + pad            in
  let y2 = y1 + pad            in
  let y3 = y2 + bheight + 1    in
  let y4 = y3 + fl_width + 2   in
  let y5 = y4 + 2              in
  let y6 = y5 + image_julia.sy in
  let y7 = y6 + 2              in
  let y8 = y7 + fl_width + 2   in
  let y9 = y8 + oheight + pad  in

  wx := x9-x0;
  wy := y9-y0+4;
  open_graph (sprintf " %dx%d" !wx !wy);
  set_font text_font;
  set_color fond_gris;
  fill_rect 0 0 !wx !wy;

  let fonglets = {fx=x1; fy=y1; fw=x8-x1; fh=y9-y1}
  and fdraw    = {fx=x4; fy=y5; fw=x5-x4; fh=y6-y5}
  and fmsg     = {fx=x2; fy=y2; fw=x7-x2; fh=bheight} in
  let onglets = make_onglets fonglets
		  ["Paramtres"; "Mandelbrot"; "Julia"; "Cache"] in

  let data = {julia   = image_julia;
	      mandel  = image_mandelbrot;
	      mem     = [];
	      maxmem  = !maxmem;
	      ongl    = onglets;
	      fongl   = fonglets;
	      fdraw   = fdraw;
	      fmsg    = fmsg;
	      bctrl   = [];
	      tat    = Done;
	      bx      = 0;
	      by      = 0;
	      fname   = ""
	     } in

  let m = Master.init () (job data) (result data) soi soj sor in

  if !port <> 0 then Master.listen m !port;
  List.iter
    (fun s -> try Master.connect m s
              with _ -> log (sprintf "unable to connect to %s\n" s))
    !slaves;
  if !auto then Master.autonomous m calc;

  let encore = ref(true) in
  sel_onglet data "Julia";
  while !encore do
    draw_onglets data.fongl data.ongl;
    let evt = match (List.hd data.ongl).titre with
      | "Paramtres" -> edit_param data
      | "Mandelbrot" -> edit_image m data MANDELBROT
      | "Julia"      -> edit_image m data JULIA
      | "Cache"      -> visu_cache data
      | _            -> failwith "cas impossible"
    in
    match evt with
      | Some(BOUTON "quitter") -> encore := false;
      | Some(ONGLET o)         -> sel_onglet data o
      | Some(RESIZE)           -> resize data
      | _ -> ()
  done;

  Master.stop m;
  exit(0)


                               (* +-----------+
                                  |  Esclave  |
                                  +-----------+ *)

let esclave() =

  let maitre = ref ""
  and port   = ref(0)
  and help   = ref false in

  let argspec = [
    "-h", Arg.Set(help),           "liste des options";
    "-m", Arg.String((:=) maitre), "<string:int> matre  contacter";
    "-p", Arg.Int((:=) port),      "<int> port TCP";
    "-v", Arg.Set(debug),          "trace les communications"
  ]
  and anon(s) = raise (Arg.Bad ("paramtre non reconnu : " ^ s))
  and msg     = "syntaxe : julia esclave <paramtres>" in

  Arg.current := 1;
  Arg.parse argspec anon msg;
  
  if !help then begin
    Arg.usage argspec msg;
    exit(1)
  end;

  let make_calc () s = calc  in
  if !maitre <> ""   then Slave.connect !maitre soi soj sor make_calc
  else if !port <> 0 then Slave.listen  !port   soi soj sor make_calc
                     else Slave.inetd           soi soj sor make_calc;
  exit(0)
;;

                         (* +-----------------------+
                            |  Programme principal  |
                            +-----------------------+ *)


begin

  match Array.to_list Sys.argv with
    | _ :: "maitre"   :: _ -> maitre()
    | _ :: "esclave"  :: _ -> esclave()
    | _ ->
	fprintf stderr "syntaxe : julia maitre   <paramtres>\n";
	fprintf stderr "          julia esclave  <paramtres>\n";
	fprintf stderr "julia (maitre|esclave) -h dtaille les paramtres\n";
	flush stderr;
	exit(1)

end
