(* +------------------------------------------------------------------------+
   |                                                                        |
   |        Recherche parallle du premier nombre premier  N chifres       |
   |                                                                        |
   | module de calcul                                                       |
   +------------------------------------------------------------------------+ *)


open Printf
open Num


              (* +---------------------------------------------+
                 |  Produit des nombres premiers impairs <= n  |
                 +---------------------------------------------+ *)

let pprime(n) =

  (* crible d'Eratosthne : t.(i) <=> 2i+3 est premier *)
  let l = (n-1)/2               in
  let t = Array.create l true   in
  let i = ref(0) and j = ref(3) in
  while !j < l do
    t.(!j) <- false;
    j := !j + !i*2 + 3;
    if !j >= l then begin
      incr i;
      while !i < l & not t.(!i) do incr i done;
      j := !i*2*(!i+3)+3
    end
  done;
  
  (* effectue le produit *)
  let p = ref(Int 1) in
  for i=0 to l-1 do if t.(i) then p := !p */ (Int(2*i+3)) done;
  !p
;;

                   (* +------------------------+
                      |  Algorithme d'Euclide  |
                      +------------------------+ *)

let rec euclide a b = if b =/ Int(0) then a else euclide b (mod_num a b);;



     (* +---------------------------------------------------------------+
        |  Cherche le premier nombre suprieur  n+k et premier avec p  |
        +---------------------------------------------------------------+ *)

let rec suivant op n p k =
  if euclide (op n (Int k)) p = Int(1) then k else suivant op n p (k+2)


                   (* +------------------------+
                      |  Test de Rabin-Miller  |
                      +------------------------+ *)

let rabin op try_giveup n0 (k,a) =

  let n  = op n0 (Int k) in
  let n1 = n -/ Int(1)  in

  (* factorise n-1 en 2^k*q *)
  let q = ref(n1)
  and i = ref(0) in
  while (mod_num !q (Int 2)) =/ (Int 0) do
    q := quo_num !q (Int 2);
    incr i
  done;

  (* dcompose q en base 2 *)
  let l = ref([]) in
  while !q >/ Int 0 do
    l := ((mod_num !q (Int 2)) <>/ (Int 0)) :: !l;
    q := quo_num !q (Int 2)
  done;

  (* calcule a^q (mod n) *)
  let p = ref(Int 1) in
  while !l <> [] do
    try_giveup();
    p := mod_num (!p */ !p) n;
    if List.hd(!l) then p := mod_num (!p */ (Int a)) n;
    l := List.tl(!l)
  done;
  
  (* lve au carr jusqu' trouver 1 ou -1 *)
  if !p =/ Int(1) then true

  else begin
    while (!p <>/ Int(1)) & (!p <>/ n1) & (!i > 1) do
      try_giveup();
      p := mod_num (!p */ !p) n;
      decr i
    done;
    !p =/ n1
  end


                      (* +----------------------------+
                         |  Distribution des travaux  |
                         +----------------------------+ *)

(*
 * stratgie : le matre slectionne les k pour lesquels n0+k est premier
 *  p et confie aux esclaves le test de Rabin sur n0+k avec la base 2.
 * Si le test est positif alors il lance le test de Rabin sur n0+k avec toutes
 * les bases prvues.
 *)


type job = int * int;;                (* (k,a) => tester Rabin (n0+k) a   *)
									  
type data = {								  
  n0 : num;            		      (* nombre de base                   *)
  p  : num;            		      (* produit des premiers impairs     *)
  op : num -> num -> num;             (* addition ou soustraction         *)
  bases : int list;    		      (* bases de test pour Rabin-Miller  *)
  mutable knext : int; 		      (* prochain n  tester = n0 + knext *)
  mutable a_faire : job list;         (* tests  effectuer                *)
  mutable kfound  : int               (* plus petit k trouv              *)
}

let job data () =
  match data.a_faire with
    | j::t -> data.a_faire  <- t; j
    | [] when data.kfound >= 0 -> raise Parallele.Job_wait
    | [] -> let k = suivant data.op data.n0 data.p data.knext in
	    data.knext <- k + 2;
            (k,2)


                       (* +---------------------------+
                          |  Rception des rsultats  |
                          +---------------------------+ *)


(* supprime ou extrait les lments vrifiant le prdicat p *)
let rec remove p = function
  | [] -> []
  | h::t -> if p(h) then remove p t else h :: remove p t

let rec extract p = function
  | [] -> []
  | h::t -> if p(h) then h :: extract p t else extract p t


(* prdicats de slection *)
let inf k = function (x,_) -> x < k
and equ k = function (x,_) -> x = k
and sup k = function (x,_) -> x > k

(* res = rabin (n0+k) a, todo = travaux en cours *) 
(* doit retourner les travaux  interrompre      *)
let result data (k,a) res en_cours =

  let infk = List.exists (inf k) (data.a_faire @ en_cours)
  and equk = List.exists (equ k) (data.a_faire @ en_cours) in

  match res with

    | true when a = 2 & data.bases <> [] ->
	(* test (k,2) positif, essaye les autres bases *)
	data.a_faire <- data.a_faire @ (List.map (fun b -> (k,b)) data.bases);
	[]

    | true when equk ->
	(* test (k,a) positif, il reste d'autres bases en cours *)
	[]

    | true when infk ->
	(* prob. premier, d'autres k plus petits sont en cours *)
	data.kfound <- k;
	data.a_faire <- remove (sup k) data.a_faire;
	extract (sup k) en_cours

    | true ->
	(* prob. premier, recherche termine *)
	data.kfound <- k;
	data.a_faire <- [];
	en_cours

    | false when (data.kfound < 0 or infk) ->
	(* compos, pas de record ou d'autres k plus petits en cours *)
	data.a_faire <- remove (equ k) data.a_faire;
	extract (equ k) en_cours

    | false ->
	(* compos, record dfinitif *)
	data.a_faire <- [];
	en_cours


                             (* +--------------+
                                |  Impression  |
                                +--------------+ *)

let soi = function n,c   -> sprintf "%d %c" n (if c then '+' else '-')
and soj = function (k,a) -> sprintf "(%d,%d)" k a
and sor = function true  -> "true" | false -> "false"
